# Autogenerated by Astropy's setup.py on 2018-12-31 23:42:30 UTC
from __future__ import unicode_literals
import datetime

version = "2.0.11"
githash = "d044049b9811eb4ff97a7ad8684c9b3efdb02ca6"


major = 2
minor = 0
bugfix = 11

release = True
timestamp = datetime.datetime(2018, 12, 31, 23, 42, 30)
debug = False

astropy_helpers_version = "2.0.8"

try:
    from ._compiler import compiler
except ImportError:
    compiler = "unknown"

try:
    from .cython_version import cython_version
except ImportError:
    cython_version = "unknown"
